@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2016 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -21,6 +21,7 @@
 
 import functools
 
+import attr
 from PyQt5.QtCore import pyqtSlot, pyqtSignal, Qt, QObject, QEvent
 from PyQt5.QtWidgets import QApplication
 
@@ -28,8 +29,10 @@ from qutebrowser.keyinput import modeparsers, keyparser
 from qutebrowser.config import config
 from qutebrowser.commands import cmdexc, cmdutils
 from qutebrowser.utils import usertypes, log, objreg, utils
+from qutebrowser.misc import objects
 
 
+@attr.s(frozen=True)
 class KeyEvent:
 
     """A small wrapper over a QKeyEvent storing its data.
@@ -43,18 +46,13 @@ class KeyEvent:
         text: A string (QKeyEvent::text).
     """
 
-    def __init__(self, keyevent):
-        self.key = keyevent.key()
-        self.text = keyevent.text()
+    key = attr.ib()
+    text = attr.ib()
 
-    def __repr__(self):
-        return utils.get_repr(self, key=self.key, text=self.text)
-
-    def __eq__(self, other):
-        return self.key == other.key and self.text == other.text
-
-    def __hash__(self):
-        return hash((self.key, self.text))
+    @classmethod
+    def from_event(cls, event):
+        """Initialize a KeyEvent from a QKeyEvent."""
+        return cls(event.key(), event.text())
 
 
 class NotInModeError(Exception):
@@ -143,9 +141,6 @@ class ModeManager(QObject):
         self._parsers = {}
         self.mode = usertypes.KeyMode.normal
         self._releaseevents_to_pass = set()
-        self._forward_unbound_keys = config.get(
-            'input', 'forward-unbound-keys')
-        objreg.get('config').changed.connect(self.set_forward_unbound_keys)
 
     def __repr__(self):
         return utils.get_repr(self, mode=self.mode)
@@ -170,23 +165,25 @@ class ModeManager(QObject):
             event.modifiers() not in [Qt.NoModifier, Qt.ShiftModifier] or
             not event.text().strip())
 
+        forward_unbound_keys = config.val.input.forward_unbound_keys
+
         if handled:
             filter_this = True
-        elif (parser.passthrough or self._forward_unbound_keys == 'all' or
-              (self._forward_unbound_keys == 'auto' and is_non_alnum)):
+        elif (parser.passthrough or forward_unbound_keys == 'all' or
+              (forward_unbound_keys == 'auto' and is_non_alnum)):
             filter_this = False
         else:
             filter_this = True
 
         if not filter_this:
-            self._releaseevents_to_pass.add(KeyEvent(event))
+            self._releaseevents_to_pass.add(KeyEvent.from_event(event))
 
         if curmode != usertypes.KeyMode.insert:
             focus_widget = QApplication.instance().focusWidget()
-            log.modes.debug("handled: {}, forward-unbound-keys: {}, "
+            log.modes.debug("handled: {}, forward_unbound_keys: {}, "
                             "passthrough: {}, is_non_alnum: {} --> "
                             "filter: {} (focused: {!r})".format(
-                                handled, self._forward_unbound_keys,
+                                handled, forward_unbound_keys,
                                 parser.passthrough, is_non_alnum, filter_this,
                                 focus_widget))
         return filter_this
@@ -201,7 +198,7 @@ class ModeManager(QObject):
             True if event should be filtered, False otherwise.
         """
         # handle like matching KeyPress
-        keyevent = KeyEvent(event)
+        keyevent = KeyEvent.from_event(event)
         if keyevent in self._releaseevents_to_pass:
             self._releaseevents_to_pass.remove(keyevent)
             filter_this = False
@@ -265,6 +262,16 @@ class ModeManager(QObject):
             m = usertypes.KeyMode[mode]
         except KeyError:
             raise cmdexc.CommandError("Mode {} does not exist!".format(mode))
+
+        if m in [usertypes.KeyMode.hint, usertypes.KeyMode.command,
+                 usertypes.KeyMode.yesno, usertypes.KeyMode.prompt]:
+            raise cmdexc.CommandError(
+                "Mode {} can't be entered manually!".format(mode))
+        elif (m == usertypes.KeyMode.caret and
+              objects.backend == usertypes.Backend.QtWebEngine):
+            raise cmdexc.CommandError("Caret mode is not supported with "
+                                      "QtWebEngine yet.")
+
         self.enter(m, 'command')
 
     @pyqtSlot(usertypes.KeyMode, str, bool)
@@ -288,7 +295,7 @@ class ModeManager(QObject):
         log.modes.debug("Leaving mode {}{}".format(
             mode, '' if reason is None else ' (reason: {})'.format(reason)))
         # leaving a mode implies clearing keychain, see
-        # https://github.com/The-Compiler/qutebrowser/issues/1805
+        # https://github.com/qutebrowser/qutebrowser/issues/1805
         self.clear_keychain()
         self.mode = usertypes.KeyMode.normal
         self.left.emit(mode, self.mode, self._win_id)
@@ -302,12 +309,6 @@ class ModeManager(QObject):
             raise ValueError("Can't leave normal mode!")
         self.leave(self.mode, 'leave current')
 
-    @config.change_filter('input', 'forward-unbound-keys')
-    def set_forward_unbound_keys(self):
-        """Update local setting when config changed."""
-        self._forward_unbound_keys = config.get(
-            'input', 'forward-unbound-keys')
-
     def eventFilter(self, event):
         """Filter all events based on the currently set mode.
 
