@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2016 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -19,12 +19,14 @@
 
 """The tab widget used for TabbedBrowser from browser.py."""
 
-import collections
 import functools
 
-from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt, QSize, QRect, QTimer, QUrl
+import attr
+from PyQt5.QtCore import (pyqtSignal, pyqtSlot, Qt, QSize, QRect, QPoint,
+                          QTimer, QUrl)
 from PyQt5.QtWidgets import (QTabWidget, QTabBar, QSizePolicy, QCommonStyle,
-                             QStyle, QStylePainter, QStyleOptionTab)
+                             QStyle, QStylePainter, QStyleOptionTab,
+                             QStyleFactory, QWidget)
 from PyQt5.QtGui import QIcon, QPalette, QColor
 
 from qutebrowser.utils import qtutils, objreg, utils, usertypes, log
@@ -50,32 +52,29 @@ class TabWidget(QTabWidget):
 
     def __init__(self, win_id, parent=None):
         super().__init__(parent)
-        bar = TabBar(win_id)
-        self.setStyle(TabBarStyle(self.style()))
+        bar = TabBar(win_id, self)
+        self.setStyle(TabBarStyle())
         self.setTabBar(bar)
         bar.tabCloseRequested.connect(self.tabCloseRequested)
         bar.tabMoved.connect(functools.partial(
-            QTimer.singleShot, 0, self.update_tab_titles))
-        bar.currentChanged.connect(self.emit_tab_index_changed)
+            QTimer.singleShot, 0, self._update_tab_titles))
+        bar.currentChanged.connect(self._on_current_changed)
         self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
         self.setDocumentMode(True)
         self.setElideMode(Qt.ElideRight)
         self.setUsesScrollButtons(True)
         bar.setDrawBase(False)
-        self.init_config()
-        objreg.get('config').changed.connect(self.init_config)
+        self._init_config()
+        config.instance.changed.connect(self._init_config)
 
     @config.change_filter('tabs')
-    def init_config(self):
+    def _init_config(self):
         """Initialize attributes based on the config."""
-        if self is None:  # pragma: no cover
-            # WORKAROUND for PyQt 5.2
-            return
         tabbar = self.tabBar()
-        self.setMovable(config.get('tabs', 'movable'))
+        self.setMovable(True)
         self.setTabsClosable(False)
-        position = config.get('tabs', 'position')
-        selection_behavior = config.get('tabs', 'select-on-remove')
+        position = config.val.tabs.position
+        selection_behavior = config.val.tabs.select_on_remove
         self.setTabPosition(position)
         tabbar.vertical = position in [QTabWidget.West, QTabWidget.East]
         tabbar.setSelectionBehaviorOnRemove(selection_behavior)
@@ -92,6 +91,23 @@ class TabWidget(QTabWidget):
         bar.set_tab_data(idx, 'indicator-color', color)
         bar.update(bar.tabRect(idx))
 
+    def set_tab_pinned(self, tab: QWidget,
+                       pinned: bool) -> None:
+        """Set the tab status as pinned.
+
+        Args:
+            tab: The tab to pin
+            pinned: Pinned tab state to set.
+        """
+        bar = self.tabBar()
+        idx = self.indexOf(tab)
+
+        bar.set_tab_data(idx, 'pinned', pinned)
+        tab.data.pinned = pinned
+        self._update_tab_title(idx)
+
+        bar.refresh()
+
     def tab_indicator_color(self, idx):
         """Get the tab indicator color for the given index."""
         return self.tabBar().tab_indicator_color(idx)
@@ -99,25 +115,35 @@ class TabWidget(QTabWidget):
     def set_page_title(self, idx, title):
         """Set the tab title user data."""
         self.tabBar().set_tab_data(idx, 'page-title', title)
-        self.update_tab_title(idx)
+        self._update_tab_title(idx)
 
     def page_title(self, idx):
         """Get the tab title user data."""
         return self.tabBar().page_title(idx)
 
-    def update_tab_title(self, idx):
+    def _update_tab_title(self, idx):
         """Update the tab text for the given tab."""
+        tab = self.widget(idx)
         fields = self.get_tab_fields(idx)
         fields['title'] = fields['title'].replace('&', '&&')
         fields['index'] = idx + 1
 
-        fmt = config.get('tabs', 'title-format')
-        title = '' if fmt is None else fmt.format(**fields)
+        fmt = config.val.tabs.title.format
+        fmt_pinned = config.val.tabs.title.format_pinned
+
+        if tab.data.pinned:
+            title = '' if fmt_pinned is None else fmt_pinned.format(**fields)
+        else:
+            title = '' if fmt is None else fmt.format(**fields)
+
         self.tabBar().setTabText(idx, title)
 
     def get_tab_fields(self, idx):
         """Get the tab field data."""
         tab = self.widget(idx)
+        if tab is None:
+            log.misc.debug("Got None-tab in get_tab_fields!")
+
         page_title = self.page_title(idx)
 
         fields = {}
@@ -126,6 +152,7 @@ class TabWidget(QTabWidget):
         fields['title_sep'] = ' - ' if page_title else ''
         fields['perc_raw'] = tab.progress()
         fields['backend'] = objects.backend.name
+        fields['private'] = ' [Private Mode] ' if tab.private else ''
 
         if tab.load_status() == usertypes.LoadStatus.loading:
             fields['perc'] = '[{}%] '.format(tab.progress())
@@ -150,21 +177,20 @@ class TabWidget(QTabWidget):
         fields['scroll_pos'] = scroll_pos
         return fields
 
-    @config.change_filter('tabs', 'title-format')
-    def update_tab_titles(self):
+    def _update_tab_titles(self):
         """Update all texts."""
         for idx in range(self.count()):
-            self.update_tab_title(idx)
+            self._update_tab_title(idx)
 
     def tabInserted(self, idx):
         """Update titles when a tab was inserted."""
         super().tabInserted(idx)
-        self.update_tab_titles()
+        self._update_tab_titles()
 
     def tabRemoved(self, idx):
         """Update titles when a tab was removed."""
         super().tabRemoved(idx)
-        self.update_tab_titles()
+        self._update_tab_titles()
 
     def addTab(self, page, icon_or_text, text_or_empty=None):
         """Override addTab to use our own text setting logic.
@@ -225,9 +251,9 @@ class TabWidget(QTabWidget):
         return new_idx
 
     @pyqtSlot(int)
-    def emit_tab_index_changed(self, index):
+    def _on_current_changed(self, index):
         """Emit the tab_index_changed signal if the current tab changed."""
-        self.tabBar().on_change()
+        self.tabBar().on_current_changed()
         self.tab_index_changed.emit(index, self.count())
 
     def tab_url(self, idx):
@@ -259,64 +285,65 @@ class TabBar(QTabBar):
     Attributes:
         vertical: When the tab bar is currently vertical.
         win_id: The window ID this TabBar belongs to.
-        _page_fullscreen: Whether the webpage (e.g. a video) is shown
-                          fullscreen.
     """
 
     def __init__(self, win_id, parent=None):
         super().__init__(parent)
         self._win_id = win_id
-        self.setStyle(TabBarStyle(self.style()))
-        self.set_font()
-        config_obj = objreg.get('config')
-        config_obj.changed.connect(self.set_font)
+        self.setStyle(TabBarStyle())
+        self._set_font()
+        config.instance.changed.connect(self._on_config_changed)
         self.vertical = False
-        self._page_fullscreen = False
         self._auto_hide_timer = QTimer()
         self._auto_hide_timer.setSingleShot(True)
-        self._auto_hide_timer.setInterval(
-            config.get('tabs', 'show-switching-delay'))
-        self._auto_hide_timer.timeout.connect(self._tabhide)
+        self._auto_hide_timer.timeout.connect(self.maybe_hide)
+        self._on_show_switching_delay_changed()
         self.setAutoFillBackground(True)
-        self.set_colors()
-        config_obj.changed.connect(self.set_colors)
-        QTimer.singleShot(0, self._tabhide)
-        config_obj.changed.connect(self.on_tab_colors_changed)
-        config_obj.changed.connect(self.on_show_switching_delay_changed)
-        config_obj.changed.connect(self.tabs_show)
+        self._set_colors()
+        QTimer.singleShot(0, self.maybe_hide)
 
     def __repr__(self):
         return utils.get_repr(self, count=self.count())
 
-    @config.change_filter('tabs', 'show')
-    def tabs_show(self):
-        """Hide or show tab bar if needed when tabs->show got changed."""
-        self._tabhide()
+    def _current_tab(self):
+        """Get the current tab object."""
+        return self.parent().currentWidget()
 
-    @config.change_filter('tabs', 'show-switching-delay')
-    def on_show_switching_delay_changed(self):
-        """Set timer interval when tabs->show-switching-delay got changed."""
-        self._auto_hide_timer.setInterval(
-            config.get('tabs', 'show-switching-delay'))
+    @pyqtSlot(str)
+    def _on_config_changed(self, option):
+        if option == 'fonts.tabs':
+            self._set_font()
+        elif option == 'tabs.favicons.scale':
+            self._set_icon_size()
+        elif option == 'colors.tabs.bar.bg':
+            self._set_colors()
+        elif option == 'tabs.show_switching_delay':
+            self._on_show_switching_delay_changed()
+        elif option == 'tabs.show':
+            self.maybe_hide()
 
-    @pyqtSlot(bool)
-    def on_page_fullscreen_requested(self, on):
-        self._page_fullscreen = on
-        self._tabhide()
+        if option.startswith('colors.tabs.'):
+            self.update()
 
-    def on_change(self):
+    def _on_show_switching_delay_changed(self):
+        """Set timer interval when tabs.show_switching_delay got changed."""
+        self._auto_hide_timer.setInterval(config.val.tabs.show_switching_delay)
+
+    def on_current_changed(self):
         """Show tab bar when current tab got changed."""
-        show = config.get('tabs', 'show')
-        if show == 'switching' or self._page_fullscreen:
+        self.maybe_hide()  # for fullscreen tabs
+        if config.val.tabs.show == 'switching':
             self.show()
             self._auto_hide_timer.start()
 
-    def _tabhide(self):
+    @pyqtSlot()
+    def maybe_hide(self):
         """Hide the tab bar if needed."""
-        show = config.get('tabs', 'show')
+        show = config.val.tabs.show
+        tab = self._current_tab()
         if (show in ['never', 'switching'] or
                 (show == 'multiple' and self.count() == 1) or
-                self._page_fullscreen):
+                (tab and tab.data.fullscreen)):
             self.hide()
         else:
             self.show()
@@ -367,50 +394,55 @@ class TabBar(QTabBar):
         # code sets layoutDirty so it actually relayouts the tabs.
         self.setIconSize(self.iconSize())
 
-    @config.change_filter('fonts', 'tabbar')
-    def set_font(self):
+    def _set_font(self):
         """Set the tab bar font."""
-        self.setFont(config.get('fonts', 'tabbar'))
+        self.setFont(config.val.fonts.tabs)
+        self._set_icon_size()
+
+    def _set_icon_size(self):
+        """Set the tab bar favicon size."""
         size = self.fontMetrics().height() - 2
+        size *= config.val.tabs.favicons.scale
         self.setIconSize(QSize(size, size))
 
-    @config.change_filter('colors', 'tabs.bg.bar')
-    def set_colors(self):
+    def _set_colors(self):
         """Set the tab bar colors."""
         p = self.palette()
-        p.setColor(QPalette.Window, config.get('colors', 'tabs.bg.bar'))
+        p.setColor(QPalette.Window, config.val.colors.tabs.bar.bg)
         self.setPalette(p)
 
-    @pyqtSlot(str, str)
-    def on_tab_colors_changed(self, section, option):
-        """Set the tab colors."""
-        if section == 'colors' and option.startswith('tabs.'):
-            self.update()
-
     def mousePressEvent(self, e):
         """Override mousePressEvent to close tabs if configured."""
-        button = config.get('tabs', 'close-mouse-button')
+        button = config.val.tabs.close_mouse_button
         if (e.button() == Qt.RightButton and button == 'right' or
                 e.button() == Qt.MiddleButton and button == 'middle'):
+            e.accept()
             idx = self.tabAt(e.pos())
-            if idx != -1:
-                e.accept()
-                self.tabCloseRequested.emit(idx)
-                return
+            if idx == -1:
+                idx = self.currentIndex()
+            self.tabCloseRequested.emit(idx)
+            return
         super().mousePressEvent(e)
 
-    def minimumTabSizeHint(self, index):
+    def minimumTabSizeHint(self, index, ellipsis: bool = True):
         """Set the minimum tab size to indicator/icon/... text.
 
         Args:
             index: The index of the tab to get a size hint for.
-
+            ellipsis: Whether to use ellipsis to calculate width
+                     instead of the tab's text.
         Return:
-            A QSize.
+            A QSize of the smallest tab size we can make.
         """
+        text = '\u2026' if ellipsis else self.tabText(index)
+        # Don't ever shorten if text is shorter than the ellipsis
+        text_width = min(self.fontMetrics().width(text),
+                         self.fontMetrics().width(self.tabText(index)))
         icon = self.tabIcon(index)
-        padding = config.get('tabs', 'padding')
+        padding = config.val.tabs.padding
+        indicator_padding = config.val.tabs.indicator_padding
         padding_h = padding.left + padding.right
+        padding_h += indicator_padding.left + indicator_padding.right
         padding_v = padding.top + padding.bottom
         if icon.isNull():
             icon_size = QSize(0, 0)
@@ -418,16 +450,32 @@ class TabBar(QTabBar):
             extent = self.style().pixelMetric(QStyle.PM_TabBarIconSize, None,
                                               self)
             icon_size = icon.actualSize(QSize(extent, extent))
-            padding_h += self.style().pixelMetric(
-                PixelMetrics.icon_padding, None, self)
-        indicator_width = config.get('tabs', 'indicator-width')
         height = self.fontMetrics().height() + padding_v
-        width = (self.fontMetrics().width('\u2026') + icon_size.width() +
-                 padding_h + indicator_width)
+        width = (text_width + icon_size.width() +
+                 padding_h + config.val.tabs.width.indicator)
         return QSize(width, height)
 
-    def tabSizeHint(self, index):
-        """Override tabSizeHint so all tabs are the same size.
+    def _tab_total_width_pinned(self):
+        """Get the current total width of pinned tabs.
+
+        This width is calculated assuming no shortening due to ellipsis."""
+        return sum(self.minimumTabSizeHint(idx, ellipsis=False).width()
+            for idx in range(self.count())
+            if self._tab_pinned(idx))
+
+    def _pinnedCount(self) -> int:
+        """Get the number of pinned tabs."""
+        return sum(self._tab_pinned(idx) for idx in range(self.count()))
+
+    def _tab_pinned(self, index: int) -> bool:
+        """Return True if tab is pinned."""
+        try:
+            return self.tab_data(index, 'pinned')
+        except KeyError:
+            return False
+
+    def tabSizeHint(self, index: int):
+        """Override tabSizeHint to customize qb's tab size.
 
         https://wiki.python.org/moin/PyQt/Customising%20tab%20bars
 
@@ -440,7 +488,7 @@ class TabBar(QTabBar):
         minimum_size = self.minimumTabSizeHint(index)
         height = minimum_size.height()
         if self.vertical:
-            confwidth = str(config.get('tabs', 'width'))
+            confwidth = str(config.val.tabs.width.bar)
             if confwidth.endswith('%'):
                 main_window = objreg.get('main-window', scope='window',
                                          window=self._win_id)
@@ -450,53 +498,54 @@ class TabBar(QTabBar):
                 width = int(confwidth)
             size = QSize(max(minimum_size.width(), width), height)
         elif self.count() == 0:
-            # This happens on startup on OS X.
+            # This happens on startup on macOS.
             # We return it directly rather than setting `size' because we don't
             # want to ensure it's valid in this special case.
             return QSize()
-        elif self.count() * minimum_size.width() > self.width():
+        else:
+            pinned = self._tab_pinned(index)
+            no_pinned_count = self.count() - self._pinnedCount()
+            pinned_width = self._tab_total_width_pinned()
+            no_pinned_width = self.width() - pinned_width
+
+            if pinned:
+                # Give pinned tabs the minimum size they need to display their
+                # titles, let Qt handle scaling it down if we get too small.
+                width = self.minimumTabSizeHint(index, ellipsis=False).width()
+            else:
+                width = no_pinned_width / no_pinned_count
+
+            # If no_pinned_width is not divisible by no_pinned_count, add a
+            # pixel to some tabs so that there is no ugly leftover space.
+            if (no_pinned_count > 0 and
+                    index < no_pinned_width % no_pinned_count):
+                width += 1
+
             # If we don't have enough space, we return the minimum size so we
             # get scroll buttons as soon as needed.
-            size = minimum_size
-        else:
-            # If we *do* have enough space, tabs should occupy the whole window
-            # width.
-            width = self.width() / self.count()
-            # If width is not divisible by count, add a pixel to some tabs so
-            # that there is no ugly leftover space.
-            if index < self.width() % self.count():
-                width += 1
+            width = max(width, minimum_size.width())
+
             size = QSize(width, height)
         qtutils.ensure_valid(size)
         return size
 
     def paintEvent(self, _e):
         """Override paintEvent to draw the tabs like we want to."""
-        # pylint: disable=bad-config-call
-        # WORKAROUND for https://bitbucket.org/logilab/astroid/issue/104
         p = QStylePainter(self)
         selected = self.currentIndex()
         for idx in range(self.count()):
             tab = QStyleOptionTab()
             self.initStyleOption(tab, idx)
 
-            bg_parts = ['tabs', 'bg']
-            fg_parts = ['tabs', 'fg']
+            # pylint: disable=bad-config-option
+            setting = config.val.colors.tabs
+            # pylint: enable=bad-config-option
             if idx == selected:
-                bg_parts.append('selected')
-                fg_parts.append('selected')
+                setting = setting.selected
+            setting = setting.odd if (idx + 1) % 2 else setting.even
 
-            if idx % 2:
-                bg_parts.append('odd')
-                fg_parts.append('odd')
-            else:
-                bg_parts.append('even')
-                fg_parts.append('even')
-
-            bg_color = config.get('colors', '.'.join(bg_parts))
-            fg_color = config.get('colors', '.'.join(fg_parts))
-            tab.palette.setColor(QPalette.Window, bg_color)
-            tab.palette.setColor(QPalette.WindowText, fg_color)
+            tab.palette.setColor(QPalette.Window, setting.bg)
+            tab.palette.setColor(QPalette.WindowText, setting.fg)
 
             indicator_color = self.tab_indicator_color(idx)
             tab.palette.setColor(QPalette.Base, indicator_color)
@@ -509,12 +558,12 @@ class TabBar(QTabBar):
     def tabInserted(self, idx):
         """Update visibility when a tab was inserted."""
         super().tabInserted(idx)
-        self._tabhide()
+        self.maybe_hide()
 
     def tabRemoved(self, idx):
         """Update visibility when a tab was removed."""
         super().tabRemoved(idx)
-        self._tabhide()
+        self.maybe_hide()
 
     def wheelEvent(self, e):
         """Override wheelEvent to make the action configurable.
@@ -522,7 +571,7 @@ class TabBar(QTabBar):
         Args:
             e: The QWheelEvent
         """
-        if config.get('tabs', 'mousewheel-tab-switching'):
+        if config.val.tabs.mousewheel_switching:
             super().wheelEvent(e)
         else:
             tabbed_browser = objreg.get('tabbed-browser', scope='window',
@@ -530,8 +579,17 @@ class TabBar(QTabBar):
             tabbed_browser.wheelEvent(e)
 
 
-# Used by TabBarStyle._tab_layout().
-Layouts = collections.namedtuple('Layouts', ['text', 'icon', 'indicator'])
+@attr.s
+class Layouts:
+
+    """Layout information for tab.
+
+    Used by TabBarStyle._tab_layout().
+    """
+
+    text = attr.ib()
+    icon = attr.ib()
+    indicator = attr.ib()
 
 
 class TabBarStyle(QCommonStyle):
@@ -550,20 +608,14 @@ class TabBarStyle(QCommonStyle):
 
     http://stackoverflow.com/a/17294081
     https://code.google.com/p/makehuman/source/browse/trunk/makehuman/lib/qtgui.py
-
-    Attributes:
-        _style: The base/"parent" style.
     """
 
-    def __init__(self, style):
+    def __init__(self):
         """Initialize all functions we're not overriding.
 
         This simply calls the corresponding function in self._style.
-
-        Args:
-            style: The base/"parent" style.
         """
-        self._style = style
+        self._style = QStyleFactory.create('Fusion')
         for method in ['drawComplexControl', 'drawItemPixmap',
                        'generatedIconPixmap', 'hitTestComplexControl',
                        'itemPixmapRect', 'itemTextRect', 'polish', 'styleHint',
@@ -638,7 +690,7 @@ class TabBarStyle(QCommonStyle):
         elif element == QStyle.CE_TabBarTabLabel:
             if not opt.icon.isNull() and layouts.icon.isValid():
                 self._draw_icon(layouts, opt, p)
-            alignment = (config.get('tabs', 'title-alignment') |
+            alignment = (config.val.tabs.title.alignment |
                          Qt.AlignVCenter | Qt.TextHideMnemonic)
             self._style.drawItemText(p, layouts.text, alignment, opt.palette,
                                      opt.state & QStyle.State_Enabled,
@@ -685,12 +737,16 @@ class TabBarStyle(QCommonStyle):
                 log.misc.warning("Could not get layouts for tab!")
                 return QRect()
             return layouts.text
-        elif sr == QStyle.SE_TabWidgetTabBar:
+        elif sr in [QStyle.SE_TabWidgetTabBar,
+                    QStyle.SE_TabBarScrollLeftButton]:
+            # Handling SE_TabBarScrollLeftButton so the left scroll button is
+            # aligned properly. Otherwise, empty space will be shown after the
+            # last tab even though the button width is set to 0
+            #
             # Need to use super() because we also use super() to render
             # element in drawControl(); otherwise, we may get bit by
             # style differences...
-            rct = super().subElementRect(sr, opt, widget)
-            return rct
+            return super().subElementRect(sr, opt, widget)
         else:
             return self._style.subElementRect(sr, opt, widget)
 
@@ -705,10 +761,10 @@ class TabBarStyle(QCommonStyle):
             opt: QStyleOptionTab
 
         Return:
-            A Layout namedtuple with two QRects.
+            A Layout object with two QRects.
         """
-        padding = config.get('tabs', 'padding')
-        indicator_padding = config.get('tabs', 'indicator-padding')
+        padding = config.val.tabs.padding
+        indicator_padding = config.val.tabs.indicator_padding
 
         text_rect = QRect(opt.rect)
         if not text_rect.isValid():
@@ -719,7 +775,7 @@ class TabBarStyle(QCommonStyle):
         text_rect.adjust(padding.left, padding.top, -padding.right,
                          -padding.bottom)
 
-        indicator_width = config.get('tabs', 'indicator-width')
+        indicator_width = config.val.tabs.width.indicator
         if indicator_width == 0:
             indicator_rect = QRect()
         else:
@@ -762,17 +818,17 @@ class TabBarStyle(QCommonStyle):
         icon_state = (QIcon.On if opt.state & QStyle.State_Selected
                       else QIcon.Off)
         # reserve space for favicon when tab bar is vertical (issue #1968)
-        position = config.get('tabs', 'position')
+        position = config.val.tabs.position
         if (opt.icon.isNull() and
                 position in [QTabWidget.East, QTabWidget.West] and
-                config.get('tabs', 'show-favicons')):
+                config.val.tabs.favicons.show):
             tab_icon_size = icon_size
         else:
             actual_size = opt.icon.actualSize(icon_size, icon_mode, icon_state)
             tab_icon_size = QSize(
                 min(actual_size.width(), icon_size.width()),
                 min(actual_size.height(), icon_size.height()))
-        icon_rect = QRect(text_rect.left(), text_rect.top() + 1,
-                          tab_icon_size.width(), tab_icon_size.height())
+        icon_top = text_rect.center().y() + 1 - tab_icon_size.height() / 2
+        icon_rect = QRect(QPoint(text_rect.left(), icon_top), tab_icon_size)
         icon_rect = self._style.visualRect(opt.direction, opt.rect, icon_rect)
         return icon_rect
